// Netify Agent Lua Processor
// Copyright (C) 2024 eGloo Incorporated
// <http://www.egloo.ca>

#pragma once

#ifdef HAVE_LUA_HPP
#include <lua.hpp>
#else
extern "C" {
#include "lauxlib.h"
#include "lua.h"
#include "lualib.h"
}
#endif

#include <atomic>
#include <condition_variable>
#include <fstream>
#include <mutex>
#include <string>
#include <unordered_map>
#include <vector>

#include <nd-plugin.hpp>

const std::unordered_map<int, std::string> lua_errors = {
    { LUA_OK, "Success" }, { LUA_ERRRUN, "Runtime Error" },
    { LUA_ERRMEM, "Memory Allocation Error" },
    { LUA_ERRERR, "Error Handler Failure" },
    { LUA_ERRSYNTAX, "Syntax Error" },
    { LUA_YIELD, "Coroutine Yields" }
};

class nppEvent
{
public:
    nppEvent(ndPluginProcessor::Event event)
      : event(event) { }
    nppEvent(ndPluginProcessor::Event event, nd_flow_ptr &flow)
      : event(event) {
        PushFlow(flow);
    }

    void PushFlow(nd_flow_ptr &flow) {
        flows.push_back(make_pair(flow, flow->stats));
    }

    ndPluginProcessor::Event event;

    typedef std::vector<std::pair<nd_flow_ptr, ndFlowStats>> Flows;
    Flows flows;
};

class nppPlugin : public ndPluginProcessor
{
public:
    enum Format {
        FORMAT_RAW,
        FORMAT_JSON,
        FORMAT_MSGPACK,
    };

    enum Compressor {
        COMPRESSOR_NONE,
        COMPRESSOR_GZ,
    };

    nppPlugin(const std::string &tag, const ndPlugin::Params &params);
    virtual ~nppPlugin();

    virtual void *Entry(void);

    virtual void GetVersion(std::string &version);

    template <class T>
    void GetStatus(T &output) const {
        nd_dprintf("%s: TODO\n", __PRETTY_FUNCTION__);
        ndPluginProcessor::GetStatus(output);
    }

    virtual void DispatchEvent(ndPlugin::Event event,
      void *param = nullptr);

    virtual void DispatchProcessorEvent(
      ndPluginProcessor::Event event, ndFlowMap *flow_map);
    virtual void DispatchProcessorEvent(
      ndPluginProcessor::Event event, nd_flow_ptr &flow);
    virtual void DispatchProcessorEvent(
      ndPluginProcessor::Event event, ndInterfaces *interfaces);
    virtual void DispatchProcessorEvent(ndPluginProcessor::Event event,
      const std::string &iface, ndPacketStats *stats);
    virtual void DispatchProcessorEvent(
      ndPluginProcessor::Event event, ndPacketStats *stats);
    virtual void DispatchProcessorEvent(
      ndPluginProcessor::Event event, ndInstanceStatus *status);
    virtual void
    DispatchProcessorEvent(ndPluginProcessor::Event event);

    const char *LoadNextScriptChunk(size_t *length);

    virtual void DispatchSinkPayload(const std::string &target,
      const ndPlugin::Channels &channels, size_t length,
      const uint8_t *payload, uint8_t flags = DF_NONE);

protected:
    lua_State *lua;

    std::atomic<bool> reload;

    enum ScriptFunctions {
        sfNONE = 0,
        sfPROCESS_AGENT_EVENT = (1 << 0),
        sfPROCESS_FLOW_EVENT = (1 << 1),
        sfPROCESS_FLOW_MAP = (1 << 2),
        sfPROCESS_FLOW_STATS = (1 << 3),
    };

    std::atomic<uint8_t> script_functions = { sfNONE };

    bool ScriptHasFunction(const std::string &func);

    void Reload(void);

    typedef std::queue<nppEvent> EventQueue;
    EventQueue event_queue;

    std::mutex event_queue_mutex;
    std::condition_variable event_queue_cond;

    void PushFlowTable(const nd_flow_ptr &flow,
      const ndFlowStats &stats, bool stats_update = false);

    void ProcessAgentEvent(nppEvent &event);
    void ProcessFlowEvent(nppEvent &event);
    void ProcessFlowMap(nppEvent &event);
    void ProcessFlowStats(nppEvent &event);

    std::string script_filename;
    std::ifstream ifs_script;
    std::vector<uint8_t> ifs_chunk;
};
